"""Handle point-and-click element selection in a Reflex app"""

import functools
import importlib.util
import inspect
from pathlib import Path
from typing import Any
import black
import reflex as rx

from . import paths

# These components will never be selectable.
ignored_components = [
    rx.accordion.root.__self__,
    rx.accordion.icon.__self__,
    rx.accordion.item.__self__,
    rx.accordion.trigger.__self__,
    rx.alert_dialog.trigger.__self__,
    rx.hover_card.root.__self__,
    rx.hover_card.trigger.__self__,
    rx.hover_card.content.__self__,
    rx.popover.root.__self__,
    rx.popover.trigger.__self__,
    rx.popover.content.__self__,
    rx.recharts.responsive_container.__self__,
    rx.select.content.__self__,
    rx.select.group.__self__,
    rx.select.item.__self__,
    rx.select.root.__self__,
    rx.theme.__self__,
    rx.Fragment,
]
# Radix triggers that fire on_pointer_down events need to be disabled so that on_click works.
neuter_pointer_down = [
    rx.select.trigger.__self__,
]
# For components that do not support on_click at all, wrap these in a box.
box_wrap_components = [
    rx.markdown.__self__,
]
# Preserve a reference to the original create method's function.
og_component_create = rx.Component.create.__func__


class ClickSelectionState(rx.State):
    """Track the currently selected component and the code that created it."""

    selected_element: str = ""
    _selected_code: list[str] = []
    filename: str = ""
    selected_function: str = ""
    start_lineno: int
    end_lineno: int
    selected_module: str = ""
    modules: list[dict[str, Any]] = []

    @rx.var
    def code(self) -> str:
        # Black format the code
        formatted_code = black.format_str("\n".join(self._selected_code), mode=black.Mode())
        return formatted_code

    async def accept_change(self):
        """Accept the current diff."""

    async def handle_click(self, filename, function_name, lineno, end_lineno, modules):
        from reflex_ai.playground import SettingsState # Importing here to avoid circular import
        # Open the toolbar
        settings_state = await self.get_state(SettingsState)
        settings_state.is_open = True
        print("clicking")
        self.selected_element = f"{filename}:{lineno}-{end_lineno}"
        self.filename = filename
        self.selected_function = function_name
        self.start_lineno = lineno
        self.end_lineno = end_lineno
        self._selected_code = (
            Path(filename).read_text().splitlines()[lineno - 1 : end_lineno]
        )
        self.modules = modules

        if paths.tmp_root_path is not None:
            module_rel_path = Path(filename).relative_to(paths.tmp_root_path).with_suffix("")
            self.selected_module = ".".join(module_rel_path.parts)
            return

        spec = importlib.util.spec_from_file_location(
            Path(filename).stem,
            filename,
            submodule_search_locations=[paths.tmp_root_path],
        )
        module = importlib.util.module_from_spec(spec)

        self.selected_module = module.__name__


def _on_click_handler(
    props: dict[str, Any],
    filename: str,
    function_name: str,
    lineno: int,
    end_lineno: int,
    modules,
):
    """Generate on_click handler and outline props for selectable components.

    Args:
        filename: The filename of the component.
        function_name: The function name of the component.
        lineno: The starting line number of the component.
        end_lineno: The ending line number of the component

    Returns:
        A dictionary of props that includes the on_click handler and outline props.
    """
    return {
        "on_click": ClickSelectionState.handle_click(
            filename,
            function_name,
            lineno,
            end_lineno,
            modules,
        ).prevent_default.stop_propagation,
        "position": "relative",
        "::before": rx.cond(
            ClickSelectionState.selected_element == f"{filename}:{lineno}-{end_lineno}",
            {
                "content": "''",
                "position": "absolute",
                "top": "-4px",
                "left": "-4px",
                "width": "calc(100% + 8px)",
                "height": "calc(100% + 8px)",
                "backgroundColor": f"var(--select-bg)",
                "outline": f"2px dashed var(--select-outline)",
                "borderRadius": "0.35rem",
                "zIndex": "999",
                "pointerEvents": "none",
                "display": "block",
            },
            {},
        ),
    }


def component_create_override(base_paths: list[Path]) -> classmethod:
    """Generate an override for Component.create that is active for the given paths.

    Args:
        base_paths: Components originating from modules prefixed by these paths will be clickable.

    Returns:
        A replacement Component.create function that adds an on_click handler where needed.
    """
    base_paths = base_paths + ["reflex_ai_tmp"]

    @classmethod
    def _component_create_override(cls, *children, **props):
        if cls not in ignored_components:
            stack = inspect.stack()
            # Walk up the stack to find the first frame that originates from a base path.
            for i, frame in enumerate(stack):
                if "site-packages" in frame.filename:
                    continue
                # Skip the module frame.
                if frame.function == "<module>":
                    continue
                if any(frame.filename.startswith(str(p)) for p in base_paths):
                    end_idx = [frame.function for frame in stack].index("playground")
                    modules = [
                        {
                            "filename": frame.filename,
                            "function": frame.function,
                        }
                        for frame in stack[i:end_idx]
                    ]
                    if cls in box_wrap_components:
                        return rx.box(
                            og_component_create(cls, *children, **props),
                            **_on_click_handler(
                                props,
                                frame.filename,
                                frame.function,
                                frame.lineno,
                                frame.positions.end_lineno,
                                modules,
                            ),
                        )
                    props.update(
                        _on_click_handler(
                            props,
                            frame.filename,
                            frame.function,
                            frame.lineno,
                            frame.positions.end_lineno,
                            modules,
                        )
                    )
                    if cls in neuter_pointer_down:
                        props.setdefault("special_props", set()).add(
                            rx.Var.create(
                                "onPointerDown={(e) => e.preventDefault()}",
                                _var_is_string=False,
                            ),
                        )
                    break
        return og_component_create(cls, *children, **props)

    return _component_create_override


def clickable(base_paths: list[Path] | None = None):
    """A decorator helper to make all components in a given page clickable to select.

    The active selection (filename and line range) is stored in
    ClickSelectionState.selected_element

    The code for the selection is cached in ClickSelectionState._selected_code

    Args:
        base_paths: Components originating from modules prefixed by these paths will be clickable.

    Returns:
        A decorator that adds an on_click handler to the component.
    """
    if base_paths is None:
        base_paths = [Path(".").resolve()]

    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            # Override the component create method.
            rx.Component.create = component_create_override(base_paths=base_paths)
            try:
                # Call the page wrapper
                return fn(*args, **kwargs)
            finally:
                # Restore the original component create method.
                rx.Component.create = classmethod(og_component_create)

        # Preserve the original function name to avoid having to specify route in add_page.
        inner.__name__ = fn.__name__
        return inner

    return outer



"""The agent that runs on the user's local client."""

import ast
import functools
import importlib
import importlib.util
import inspect
import filecmp
import difflib
from pathlib import Path

import black
from flexai import Agent
from flexai.message import Message, ToolResultMessage
from pydantic import BaseModel
from reflex_ai import paths


# NOTE: using BaseModel here instead of rx.Base due to FastAPI not liking the v1 models
class InternRequest(BaseModel):
    """The request to the AI intern."""

    prompt: str
    selected_code: str
    selected_module: str
    selected_function: str


class InternResponse(BaseModel):
    """The response from the AI intern."""

    request_id: str
    messages: list[Message]


class ToolRequestResponse(BaseModel):
    """The response from the tool to the AI intern."""

    request_id: str
    messages: list[ToolResultMessage]


class EditResult(BaseModel):
    """The result of an edit."""

    request_id: str
    diff: str
    accepted: bool


def get_agent() -> Agent:
    """Get an instance of an intern."""
    return Agent(
        tools=[
            get_module_source,
            add_python_element,
            update_python_element,
            delete_python_element,
        ],
    )


@functools.lru_cache
def import_module_from_tmp(module_name: str):
    """Import the module from the scratchpad."""
    module = importlib.import_module(module_name)
    module_file = inspect.getsourcefile(module)
    module_tmp_file = module_file.replace(str(paths.base_paths[0].parent), str(paths.tmp_root_path))
    spec = importlib.util.spec_from_file_location(
        module.__name__,
        module_tmp_file,
    )
    new_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(new_module)
    return new_module

def diff_directories(dir1, dir2):
    """
    Compare two directories recursively. Files in each directory are
    assumed to be equal if their names and contents are equal.
    Returns a dictionary with file paths as keys and their diffs as values.
    """
    print("diffing", dir1, dir2)

    diffs = {}

    def compare_dirs(dcmp):
        for name in dcmp.diff_files:
            if not name.endswith(".py"):
                continue
            file1 = Path(dcmp.left) / name
            file2 = Path(dcmp.right) / name
            with file1.open() as f1, file2.open() as f2:
                f1_lines = f1.read().splitlines()
                f2_lines = f2.read().replace("EditableState", "rx.State").replace("from reflex_ai import rx.State\n", "").splitlines()
                diff = list(difflib.unified_diff(
                    f1_lines,
                    f2_lines,
                    fromfile=str(file1),
                    tofile=str(file2)
                ))
                diffs[file1] = diff
        for sub_dcmp in dcmp.subdirs.values():
            compare_dirs(sub_dcmp)

    dirs_cmp = filecmp.dircmp(dir1, dir2)
    compare_dirs(dirs_cmp)

    return diffs

def directory_diff():
    """Diff the scratchpad and the base directories."""
    return diff_directories(paths.base_paths[0].parent, paths.tmp_root_path)

def get_module_ast(module_name: str) -> ast.Module:
    """Get the AST of a module.

    Args:
        module_name: The name of the module.

    Returns:
        The AST of the module.
    """
    # Import the module
    module_source = get_module_source(module_name)

    # Parse the module source code into an AST
    module_ast = ast.parse(module_source)
    return module_ast


def write_module_source(module_name: str, source_code: str):
    """Write the source code of a module.

    Args:
        module_name: The name of the module.
        source_code: The source code of the module.
    """
    # Import the module.
    module = import_module_from_tmp(module_name)

    # Get the source code of the module.
    module_file = inspect.getsourcefile(module)

    # Format the source code using Black.
    source_code = black.format_str(source_code, mode=black.FileMode())

    # Write the new source code back to the module file.
    with open(module_file, "w") as file:
        file.write(source_code)


def get_module_source(module_name: str) -> str:
    """Get the source code of a module."""
    # Import the module
    module = import_module_from_tmp(module_name)

    # Get the source code of the module
    module_file = inspect.getsourcefile(module)
    with open(module_file, "r") as file:
        module_source = file.read()
    return module_source


def _find_node(tree, element_type: str, element_name: str):
    for node in ast.walk(tree):
        if (
            isinstance(node, ast.FunctionDef)
            and element_type == "function"
            and node.name == element_name
        ):
            return node
        elif (
            isinstance(node, ast.ClassDef)
            and element_type == "class"
            and node.name == element_name
        ):
            return node
        elif isinstance(node, ast.Assign) and element_type == "variable":
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == element_name:
                    return node
    return None

import astor

def update_python_element(
    module_name: str,
    element_type: str,
    element_name: str,
    new_content: str,
    class_name: str = None,
):
    """Update the source code of an element (class, function, or variable) in a module.
    
    Args:
        module_name: The name of the module.
        element_type: One of "class", "function", or "variable".
        element_name: The name of the element to update.
        new_content: The new source code of the element.
        class_name: The name of the class containing the element (if applicable).
    """
    tree = get_module_ast(module_name)

    if class_name:
        class_node = _find_node(tree, "class", class_name)
        if not class_node:
            raise ValueError(f"Class {class_name} not found")
        target_node = _find_node(
            class_node, element_type, element_name
        )
    else:
        target_node = _find_node(tree, element_type, element_name)

    if not target_node:
        raise ValueError(
            f"{element_type.capitalize()} {element_name} not found"
        )

    new_node = ast.parse(new_content).body[0]

    if element_type == "variable":
        target_node.value = new_node.value
    else:
        for i, item in enumerate(target_node.body):
            if isinstance(item, ast.Expr) and isinstance(item.value, ast.Str):
                # Preserve docstring
                new_node.body.insert(0, item)
                break
        target_node.body = new_node.body

    write_module_source(module_name, astor.to_source(tree))

def delete_python_element(
    module_name: str, element_type: str, element_name: str, class_name: str = None
) -> bool:
    """Delete an element (class, function, or variable) from a module.
    
    Args:
        module_name: The name of the module.
        element_type: One of "class", "function", or "variable".
        element_name: The name of the element to delete.
        class_name: The name of the class containing the element (if applicable).
    """
    tree = get_module_ast(module_name)

    if class_name:
        class_node = _find_node(tree, "class", class_name)
        if not class_node:
            raise ValueError(f"Class {class_name} not found")
        class_node.body = [
            node
            for node in class_node.body
            if not (
                (
                    isinstance(node, ast.FunctionDef)
                    and element_type == "function"
                    and node.name == element_name
                )
                or (
                    isinstance(node, ast.Assign)
                    and element_type == "variable"
                    and any(
                        t.id == element_name
                        for t in node.targets
                        if isinstance(t, ast.Name)
                    )
                )
            )
        ]
    else:
        tree.body = [
            node
            for node in tree.body
            if not (
                (
                    isinstance(node, ast.FunctionDef)
                    and element_type == "function"
                    and node.name == element_name
                )
                or (
                    isinstance(node, ast.ClassDef)
                    and element_type == "class"
                    and node.name == element_name
                )
                or (
                    isinstance(node, ast.Assign)
                    and element_type == "variable"
                    and any(
                        t.id == element_name
                        for t in node.targets
                        if isinstance(t, ast.Name)
                    )
                )
            )
        ]

    write_module_source(module_name, astor.to_source(tree))

def add_python_element(
    module_name: str,
    content: str,
    class_name: str = None,
    position: str = "end",
) -> bool:
    """Add a new element (class, function, or variable) to a Python file.
    
    Args:
        filename: The name of the file to add the element to.
        content: The source code of the element.
        class_name: The name of the class containing the element (if applicable).
        position: The position to add the element in the module. One of "start", "end", or "after_import".
    """
    tree = get_module_ast(module_name)
    new_element = ast.parse(content).body[0]

    if class_name:
        class_node = _find_node(tree, "class", class_name)
        if not class_node:
            raise ValueError(f"Class {class_name} not found")
        target_body = class_node.body
    else:
        target_body = tree.body

    if position == "start":
        target_body.insert(0, new_element)
    elif position == "end":
        target_body.append(new_element)
    else:
        # Find the position after the last import statement
        insert_position = 0
        for i, node in enumerate(target_body):
            if isinstance(node, (ast.Import, ast.ImportFrom)):
                insert_position = i + 1
            else:
                break
        target_body.insert(insert_position, new_element)
    write_module_source(module_name, astor.to_source(tree))



"""The local AI toolbar to hook up with the server-side reflex agent."""

import json
import os

import httpx
import reflex as rx
from anthropic.types import ToolUseBlock
from flexai.message import Message
from reflex_ai.selection import ClickSelectionState
from reflex_ai.local_agent import (
    get_agent,
    InternRequest,
    InternResponse,
    ToolRequestResponse,
    EditResult,
    directory_diff,
)
from reflex_ai import utils, paths

# Tools mapping to human readable names
tools_hr_names = {
    "get_module_source": "Analyzing code",
    "add_python_element": "Adding new elements",
    "update_python_element": "Updating elements",
    "delete_python_element": "Deleting elements",
}


async def make_request(
    endpoint: str,
    data: dict,
    url: str = os.getenv("FLEXGEN_BACKEND_URL", "http://localhost:8000"),
    timeout: int = 60,
) -> dict:
    """Make a request to the backend.

    Args:
        endpoint: The endpoint to send the request to.
        data: The data to send.
        url: The URL of the backend.
        timeout: The timeout for the request.

    Returns:
        The JSON response from the backend.
    """
    async with httpx.AsyncClient() as client:
        resp = await client.post(
            f"{url}/api/{endpoint}",
            data=data,
            timeout=timeout,
        )

    print(resp)
    print(resp.json())
    resp.raise_for_status()
    return resp.json()


class Diff(rx.Base):
    filename: str
    diff: str

class ToolbarState(rx.State):
    """The toolbar state."""

    tools_used: list[str] = ["Preparing request"]
    current_step: str = "selected_code" # The current step name of the process (selected_code, processing, review_changes)
    processing: bool = False
    selected_id: str = ""
    code: str = ""
    prompt: str = ""
    step: int = 0
    diff: list[Diff] = []
    selected_diff: Diff = Diff(filename="", diff="") # The selected diff
    changes_comment: str = ""
    edit_id: str = ""

    async def process(self, prompt: dict[str, str]):
        """Process the user's prompt.

        Args:
            prompt: The prompt from the user from the form input.
        """
        # Start the processing.
        self.start_processing()
        self.prompt = prompt["prompt"]
        yield

        # Get the selected code.
        selection_state = await self.get_state(ClickSelectionState)
        selected_code = "\n".join(selection_state._selected_code)

        # Create the intern request.
        request = InternRequest(
            prompt=prompt["prompt"],
            selected_code=selected_code,
            selected_module=selection_state.selected_module,
            selected_function=selection_state.selected_function,
        )
        response = await make_request("intern", request.model_dump_json())
        resp_obj = InternResponse(**response)
        messages = [Message(role=m.role, content=m.content) for m in resp_obj.messages]

        # Process the messages with the local agent.
        local_intern = get_agent()
        unconverted_messages = []

        # Hack until we have state maintain between hot reloads.
        utils.save_request_id(resp_obj.request_id)

        # Run in a loop until we're done with the request.
        while True:
            # Get any tool use messages from the intern and process them.
            tool_response_messages = []
            for message in messages:
                try:
                    tool_use_message = local_intern.llm.to_tool_use_message(
                        ToolUseBlock.parse_raw(message.content),
                    )
                except ValueError:
                    unconverted_messages.append(message)
                    continue
                self.tools_used.append(tools_hr_names.get(tool_use_message.tool_name, tool_use_message.tool_name))
                yield
                # Invoke the tool and get the response.
                response = await local_intern.invoke_tool(tool_use_message)

                if tool_use_message.tool_name == "update_python_element":
                    import importlib.util
                    parent = self.get_value(selection_state.modules)[0]
                    filename = parent["filename"]

                    # Load the module from the filename
                    module_name = os.path.splitext(os.path.basename(filename))[0]
                    spec = importlib.util.spec_from_file_location(module_name, filename)
                    module = importlib.util.module_from_spec(spec)
                    os.environ["PYTEST_CURRENT_TEST"] = "True"
                    spec.loader.exec_module(module)
                    component_fn = getattr(module, parent["function"])
                    try:
                        component_fn()
                    except Exception as e:
                        from flexai.message import ToolResultMessage
                        response = ToolResultMessage.from_result(
                            tool_use_message.id, e, execution_time=0,
                        )
                        print("hit an error", e)

                tool_response_messages.append(response)
                print("response", tool_response_messages[-1])
                # Diff the directories.
                self.load_diff()

            # Base case: no more messages to process.
            if not tool_response_messages:
                break

            # Send the tool response to the intern.
            tool_response_request = ToolRequestResponse(
                request_id=resp_obj.request_id,
                messages=tool_response_messages,
            )
            response = await make_request(
                "intern/tool_response", tool_response_request.model_dump_json()
            )
            messages = [Message(**m) for m in response]

        self.finish_processing(messages)
        yield
        # Touch the rxconfig.py to trigger a hot reload.
        self.trigger_reload()

    def start_processing(self):
        self.tools_used = [self.tools_used[0]]
        self.processing = True
        self.current_step = "processing"
        self.step = 1

    def finish_processing(self, messages):
        self.selected_diff = self.diff[0]
        self.changes_comment = messages[-1].content
        self.current_step = "review_changes"
        self.step = 2
        self.processing = False

    async def accept_change(self):
        """Accept the current diff."""
        print("Accepting changes.")
        request_id = utils.load_request_id()
        await make_request("intern/edit_result", data=EditResult(request_id=request_id, diff=json.dumps([d.dict() for d in self.diff]), accepted=True).model_dump_json())
        print("sent request")
        utils.commit_scratch_dir(paths.base_paths[0], [d.filename for d in self.diff])

    async def revert_change(self):
        """Revert the current diff."""
        # Rewrite the scratch directory.
        print("Reverting changes.")
        request_id = utils.load_request_id()
        await make_request("intern/edit_result", data=EditResult(request_id=request_id, diff=json.dumps([d.dict() for d in self.diff]), accepted=False).model_dump_json())
        utils.create_scratch_dir(paths.base_paths[0], overwrite=True)
        print("sent request")
        self.trigger_reload()

    def trigger_reload(self):
        """Trigger a hot reload."""
        contents = open("rxconfig.py").read()
        with open("rxconfig.py", "w") as f:
            f.write(contents)

    def load_diff(self):
        diff = directory_diff()
        self.diff = [Diff(filename=str(filename), diff="\n".join(diff)) for filename, diff in diff.items()]


"""Utility functions for the reflex_ai package."""

import os
import shutil
from pathlib import Path


SCRATCH_DIR_NAME = "reflex_ai_tmp"


def create_scratch_dir(app_dir: Path, overwrite: bool = False) -> Path:
    """Create a scratch directory for the agent to make changes to.

    Args:
        app_dir: The directory of the app to copy into the scratch directory.
        overwrite: Whether to overwrite the scratch directory if it already exists.

    Returns:
        The path to the created directory.
    """
    print("creating scratch dir", app_dir)
    scratch_dir = app_dir / SCRATCH_DIR_NAME

    # If the scratch directory already exists, skip.
    if scratch_dir.exists() and not overwrite:
        return scratch_dir

    # Copy the app directory to a temporary path for modifications.
    shutil.copytree(
        app_dir.parent,
        scratch_dir,
        ignore=lambda _, names: [scratch_dir.name, ".web", "assets", "__pycache__"],
        dirs_exist_ok=True,
    )

    # Perform search and replace on the copied files
    for root, _, files in os.walk(scratch_dir):
        print(files)
        for file in files:
            if not file.endswith(".py") or "rxconfig" in file:
                continue
            # Construct full file path
            file_path = os.path.join(root, file)

            # Read the file content
            with open(file_path, "r") as f:
                content = f.read()

            # Perform replacements
            content = content.replace("(rx.State)", "(EditableState)")
            content = f"from reflex_ai import EditableState\n{content}"

            # Write the modified content to the file
            print("writing to", file_path)
            with open(file_path, "w") as f:
                f.write(content)

    return scratch_dir


def commit_scratch_dir(app_dir: Path, files: list[str]):
    """Copy all files from the scratch directory back to the corresponding app directory.

    Args:
        app_dir: The original app directory to copy files back to.
        files: The list of files to copy back.
    """
    scratch_dir = app_dir / SCRATCH_DIR_NAME
    for file in files:
        # Construct corresponding file path in app directory
        print("checking file", file)
        relative_path = os.path.relpath(file, app_dir.parent)
        # Copy the file
        # Read the file content
        with open(scratch_dir / relative_path, "r") as f:
            content = f.read()

        # Perform replacements
        content = content.replace("(EditableState)", "(rx.State)")
        content = content.replace("from reflex_ai import EditableState\n", "")

        # Write the modified content to the file
        print("writing to", file)
        with open(file, "w") as f:
            f.write(content)

        print(f"Copied {file} to {file}")

def save_request_id(request_id: str):
    """Save the request ID to the request ID file."""
    with open("request_id.txt", "w") as f:
        f.write(request_id)

def load_request_id() -> str:
    """Load the request ID from the request ID file."""
    with open("request_id.txt", "r") as f:
        return f.read()
